#include <stdlib.h>
#include <stdio.h>
#include <string.h>

const double load_factor = 0.5;

typedef struct str_node {
  char *str;
  int val;
  struct str_node *next;
} str_node;

typedef struct hash_table {
  int size;
  int count;
  str_node **a;
} hash_table;

int hash_string(char *s) {
  int i, r = 0;

  for (i = 0; s[i] != 0; i++) {
    r += s[i];
  }

  return r;
}

int better_hash_string(char *s) {
  int i, r = 0;

  for (i = 0; s[i] != 0; i++) {
    r *= 10;
    r += s[i];
  }

  if (r < 0) r = -r;

  return r;
}

void resize_hash_table(hash_table *ht);

void reset_hash_table(hash_table *ht, int size)
{
  int i;

  ht->size = size;
  ht->count = 0;

  ht->a = (str_node**)malloc(ht->size * sizeof(str_node*));
  for (i = 0; i < size; i++)
    ht->a[i] = NULL;
}

hash_table* make_hash_table()
{
  hash_table *ht;

  ht = (hash_table *)malloc(sizeof(hash_table));
  reset_hash_table(ht, 8);
  
  return ht;
}

int hash_get(hash_table *ht, char *s)
{
  int n = hash_string(s) % ht->size;
  str_node *nd = ht->a[n];

  while (nd != NULL) {
    if (!strcmp(nd->str, s))
      return nd->val;
    nd = nd->next;
  }
  
  return -1;
}

void hash_set(hash_table *ht, char *s, int val)
{
  int n = hash_string(s) % ht->size;
  str_node *nd;

  nd = ht->a[n];
  while (nd != NULL) {
    if (!strcmp(nd->str, s)) {
      nd->val = val;
      return;
    }
    nd = nd->next;
  }

  nd = (str_node*)malloc(sizeof(str_node));
  nd->str = s;
  nd->val = val;
  nd->next = ht->a[n];

  ht->a[n] = nd;

  ht->count++;

  if (ht->count > ht->size * load_factor)
    resize_hash_table(ht);
}

void resize_hash_table(hash_table *ht) {
  int i, size = ht->size;
  str_node **a = ht->a, *nd;
  
  reset_hash_table(ht, size * 2);
  for (i = 0; i < size; i++) {
    nd = a[i];
    while (nd != NULL) {
      hash_set(ht, nd->str, nd->val);
      nd = nd->next;
    }
  }
}

int main(int argc, char** argv)
{
  hash_table* ht;
  char buffer[256];
  int i;
  FILE *f;

  ht = make_hash_table();

  f = fopen(argv[1], "r");
  while (fgets(buffer, 256, f)) {
    hash_set(ht, strdup(buffer), 1);
  }
  fclose(f);

  printf("%d words\n", ht->count);

  while (fgets(buffer, 256, stdin)) {
    i = hash_get(ht, buffer);
    if (i < 0)
      printf("Not there\n");
    else
      printf("Value: %d\n", i);
  }

  return 0;
}
